%option pointer
%option noyywrap
%option nounput
%option nostack
%option nomain
%option never-interactive
%option noyylineno

%{
#include <unistd.h>
#include "sml.h"
#include "stralloc.h"
#include "strerr.h"
#include "lexer.h"
#include "alloc.h"
#include "gen_alloc.h"
#include "gen_allocdefs.h"

/*--------------------------------------------------------------------*/
/* Flex things */
#ifdef YY_INPUT
#undef YY_INPUT
#endif
#define YY_INPUT(buf,result,max_size) result = read_and_copy(buf,max_size)

static int
read_and_copy(char *buf, int max_size)
{
    int rc = read(0, buf, max_size);
    if (rc < 0)
	strerr_die1sys(1, "read: ");

    if (!stralloc_catb(safe_output, buf, rc))
	strerr_die1sys(1, "saving input: ");

    return rc;
}

#ifdef YY_FATAL_ERROR
#undef YY_FATAL_ERROR
#endif
#define YY_FATAL_ERROR(m) strerr_die1x(1, m)

/* Shut up gcc */
void
shut_up_yy_fatal_error_unused_warning(void)
{
    yy_fatal_error(0);
}

#define yywrap() 1

#ifdef YY_READ_BUF_SIZE
#undef YY_READ_BUF_SIZE
#endif
#define YY_READ_BUF_SIZE 262144

#ifdef YY_BUF_SIZE
#undef YY_BUF_SIZE
#endif
#define YY_BUF_SIZE 262144

/*--------------------------------------------------------------------*/
/* Private Context stack used only in lexer. */

GEN_ALLOC_typedef(Stack, Context, s, len, a)

static
GEN_ALLOC_readyplus(Stack, Context, s, len, a, i, n, x, 30, Stack_readyplus)

static Stack stack;

static Context
pop(void)
{
    if (!stack.s || !stack.len)
	strerr_die1x(1, "illegal pop: stack empty");

    return stack.s[--stack.len];
}

static void
push(Context x)
{
    if (!Stack_readyplus(&stack, 1))
	strerr_die1sys(1, "pushing context: ");
    stack.s[stack.len++] = x;
}

/*--------------------------------------------------------------------*/
/* Private state & utilities */

static Context x;
static Run *lastrun;
static char lastc;

static void
do_B(void)
{
    push(x);
    x.flags |= B;
}

static void
do_EM(void)
{
    push(x);
    if (!(x.flags & S)) x.flags ^= EM;
}

static void
do_I(void)
{
    push(x);
    x.flags |= I;
}

static void
do_PL(void)
{
    push(x);
    x.flags = 0;
}

static void
do_S(void)
{
    push(x);
    x.flags |= S;
    x.flags &= ~EM;
}

static void
do_TT(void)
{
    push(x);
    x.flags |= TT;
}

static void
do_U(void)
{
    push(x);
    if ((x.flags & U) < 3) x.flags++;
}

static void
do_size(char size)
{
    push(x);
    x.size = size;
}

static void
do_color(char color)
{
    push(x);
    x.color = color;
}

static void
do_close(void)
{
    x = pop();
}

static void
new_run(Context x)
{
    Run run;
    run.len = 0;
    run.x = x;
    if (!Runs_append(&markup, &run))
	strerr_die1sys(1, "appending to markup: ");
    lastrun = markup.s + (markup.len - 1);
}

static void
append(char c, Context x)
{
    if (!stralloc_append(&chars, &c)) {
	strerr_die1sys(1, "appending to chars: ");
    }
    if (Context_diff(lastrun->x, x)) new_run(x);
    lastrun->len++;
    lastc = c;
}

static void
do_space(char c)
{
    Context sx = to_space_context(x, lastrun->x);

    if (x.flags & TT) {
	append(c, sx);
    }

    else if (lastc != ' ' || Context_diff(lastrun->x, sx)) {
	append(' ', sx);
    }
}

static void
die_badchar(char c)
{
    strerr_die1x(1, "illegal character");
}

static void
finish(void)
{
    if (stack.len > 0)
	strerr_die1x(1, "unclosed tags at EOF");

    alloc_free(stack.s);
    stack.s = 0;
}

%}

%%
%{
    new_run(x);
%}

"<B>"				do_B();
"<EM>"				do_EM();
"<I>"				do_I();
"<S>"				do_S();
"<TT>"				do_TT();
"<PL>"				do_PL();
"</"(B|EM|I|S|TT|PL|U)\>	do_close();
"<U>"				do_U();
\<[0-9]\>			do_size(yytext[1]);
\<[rgbcmykw]\>			do_color(yytext[1]);
"</"[0-9rgbcmykw]\>		do_close();
[<>]				die_badchar(yytext[0]);
[ \t\r\n]			do_space(yytext[0]);
[\041-\176]			append(yytext[0], x);
.				die_badchar(yytext[0]);
<<EOF>>				finish(); return;
%%
